#!/usr/bin/env python
#
# @package      hubzero-forge
# @file         gengitExternalapache.py
# @copyright    Copyright (c) 2010-2020 The Regents of the University of California.
# @license      http://opensource.org/licenses/MIT MIT
#
# Copyright (c) 2010-2020 The Regents of the University of California.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# HUBzero is a registered trademark of The Regents of the University of California.
#

# Really quick and dirty port of the perl script of similar name
# but querying project data from MySQL instead of LDAP.
# This should get cleaned up an rewritten to use
# the hubzero python library (which isn't ready
# at the time of this writing).
#

import os
import sys
import argparse
import re
import subprocess
try:
   import configparser
except:
   import ConfigParser as configparser
try:
   import pymysql as mysqldb
except:
   import MySQLdb as mysqldb

#
# @TODO: Add commandline or environment variable overrides to the autodetection
#

parser = argparse.ArgumentParser(description="Create the Apache configuration file controlling access to GIT repositories.")
parser.add_argument('--gitroot',required=True,help="path to GIT repositories")
parser.add_argument('--giturl',required=True,help="URL path to GIT repository, relative to the document root")
parser.add_argument('--gitconf',required=True,help="path to Apache configuration file")
parser.add_argument('--projectType',required=False,help="project type")

args = parser.parse_args()

#print(args)

gitroot     = args.gitroot
giturl      = args.giturl
gitconf     = args.gitconf
projectType = args.projectType

if not os.path.isdir(gitroot):
   sys.stderr.write("Error:  '" + gitroot + "' is not a valid directory.\n")
   sys.exit(2)

#
# Detect whether we are running Apache 2.2 or not so we set proper configuration
# directives later on. We assume 2.2 until proven otherwise. We also assume
# standard Debian or Redhat server locations. If running a custom installation
# this will likely fall back to assuming 2.2.
#
apache_2_2 = True
apacheBin = False

apachePath = os.path.join(os.sep,'usr','sbin','apache2')
if os.path.exists(apachePath):
   apacheBin = apachePath
else:
   apachePath = os.path.join(os.sep,'usr','sbin','httpd')
   if os.path.exists(apachePath):
      apacheBin = apachePath

if apacheBin:
   proc = subprocess.Popen([apacheBin, "-v"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
   out,err = proc.communicate()

   apacheVersion = re.compile(b'.*Apache/(\d+)\.(\d+)\.(\d+)',flags=re.MULTILINE)

   m = apacheVersion.match(out)

   if m and int(m.group(1)) >= 2 and int(m.group(2)) > 2:
      apache_2_2 = False


def loadHubConfigurationData(configurationFileDirname,
                             configurationFileiBasename):
   hubconfigurationData = {}
   if   not os.path.isdir(configurationFileDirname):
      sys.stderr.write("ERROR: specified base directory does not exist at %s\n" % (configurationFileDirname))
      sys.exit(5)
   else:
      hubconfigurationPath = os.path.join(configurationFileDirname,configurationFileiBasename)
      if not os.path.isfile(hubconfigurationPath):
         sys.stderr.write("ERROR: specified hubconfiguration.php file does exist at %s\n" % (hubconfigurationPath))
         sys.exit(5)
      else:
         reHubconfigurationVar = re.compile("\s*(?:var|public)\s+\$(\w+)\s*=\s*\'*(.*?)\'*\s*\;\s*")
         try:
            with open(hubconfigurationPath,'r') as fpHubconfiguration:
               for line in fpHubconfiguration:
                  match = reHubconfigurationVar.match(line)
                  if match:
                     hubconfigurationData[match.group(1)] = match.group(2).strip(" \'\"\t").replace("\\'","'")
         except:
            sys.stderr.write("ERROR: specified hubconfiguration file %s could not be ingested\n" % (hubconfigurationPath))
            sys.stderr.write(traceback.format_exc())
            sys.exit(5)

   return(hubconfigurationData)


config = configparser.RawConfigParser()
config.read(os.path.join(os.sep,'etc','hubzero.conf'))
section = config.get('DEFAULT','site')
docroot = config.get(section,'DocumentRoot')

data = {}
data['joomla']  = loadHubConfigurationData(docroot,'configuration.php')
data['hubzero'] = loadHubConfigurationData(docroot,'hubconfiguration.php')

mysql_host     = data['joomla']['host']
mysql_user     = data['joomla']['user']
mysql_password = data['joomla']['password']
mysql_db       = data['joomla']['db']

try:
   db = mysqldb.connect(host=mysql_host, user=mysql_user, passwd=mysql_password, db=mysql_db)
except:
   sys.stderr.write("Error: Unable to connect to HUBzero database.\n")
   sys.exit(4)
else:
   cursor = db.cursor()
   if not cursor:
      sys.stderr.write("Error: Unable to obtain HUBzero database cursor.\n")
      sys.exit(5)

ldapbase     = data['hubzero']['hubLDAPBaseDN']
ldaphost     = data['hubzero']['hubLDAPMasterHost']
ldapbinduser = data['hubzero']['hubLDAPSearchUserDN']
ldapbindpass = data['hubzero']['hubLDAPSearchUserPW']
try:
   ldapgroupobjectclass = data['hubzero']['hubLDAPGroupObjectClass']
except:
   ldapgroupobjectclass = 'posixGroup'
finally:
   if ldapgroupobjectclass != 'posixGroup':
      ldapgroupobjectclass = 'hubGroup'

gitCorePath = os.path.join(os.sep,'usr','lib','git-core')
if not os.path.exists(gitCorePath):
   gitCorePath = os.path.join(os.sep,'usr','libexec','git-core')
   if not os.path.exists(gitCorePath):
      sys.stderr.write("ERROR: git-core path could not be determined.\n")
      sys.exit(3)

try:
   fpGitConf = open(gitconf,'w')
except:
   sys.stderr.write("ERROR: Attempt to open %s failed.\n" % (gitconf))
else:
   apacheHeader = """#
# Apache configuration for GIT repositories
#
# This file is generated automatically from MySQL.
#      !!!  DO NOT MODIFY THIS FILE  !!!
#

SetEnvIf Request_URI "(?x)^/tools/([^/]+)/gitExternal/\\1/" GIT_PROJECT_ROOT=/opt/gitExternal/tools
SetEnv GIT_HTTP_EXPORT_ALL

# Matching pattern /tools/toolname/git/toolname
# Putting toolname at the end of the URL means that by
# default the git clone command will put the repostory
# in a directory named toolname
#
# The command
#     git clone https://ebbtide.aws.hubzero.org/tools/toolname/git/toolname
# generates these requests
#     GET /tools/toolname/git/toolname/info/refs?service=git-upload-pack HTTP/1.1
#     GET /tools/toolname/git/toolname/info/refs?service=git-upload-pack HTTP/1.1
#     GET /tools/toolname/git/toolname/info/refs?service=git-upload-pack HTTP/1.1
#     POST /tools/toolname/git/toolname/git-upload-pack HTTP/1.1

ScriptAliasMatch \\
    "(?x)^/tools/([^/]+)/gitExternal/\\1/(.*?(HEAD | \\
                                             info/refs | \\
                                             objects/(info/[^/]+ | \\
                                                      [0-9a-f]{2}/[0-9a-f]{38} | \\
                                                      pack/pack-[0-9a-f]{40}\.(pack|idx)) | \\
                                             git-(upload|receive)-pack))$" \\
    %s/git-http-backend/$1/$2

RewriteCond %%{QUERY_STRING} service=git-receive-pack [OR]
RewriteCond %%{REQUEST_URI} /git-receive-pack$
RewriteRule ^ - [E=AUTHREQUIRED:yes]

""" % (gitCorePath)

   fpGitConf.write(apacheHeader)

   for dirEntry in os.listdir(gitroot):
      if not os.path.isdir(os.path.join(gitroot,dirEntry)):
         continue

      sourcePublic = False
      id = dirEntry.split('.git')[0]
      if projectType != 'tool':
         gid = []
         gid.append(id)
         gid.append('apps')
         name = id
      else:
         gid = set([])
         cursor.execute("SELECT title,codeaccess,tg.cn FROM jos_tool_version AS tv, jos_tool_groups AS tg WHERE tg.toolid=tv.toolid AND tv.toolname='" + id + "' AND state='3' AND (tg.role=1 or tg.role=2)")
         rows = cursor.fetchall()
         if len(rows) == 0:
            continue
         name = id
         for name,sourceOpen,gidref in rows:
            if sourceOpen == '@OPEN':
               sourcePublic = True
            else:
               sourcePublic = False
            if gidref == 'apps':
               gid.add(gidref)

# repository is available at external site
      sourcePublic = False
      url = giturl.replace("@PROJECT@",id)

      fpGitConf.write("<LocationMatch %s>\n" % (url))
      fpGitConf.write("\tOrder Deny,Allow\n")
      if sourcePublic:
         fpGitConf.write("\tDeny from env=AUTHREQUIRED\n")
      else:
         fpGitConf.write("\tDeny from All\n")
      fpGitConf.write("\tAuthType Basic\n")
      fpGitConf.write("\tAuthBasicProvider ldap\n")
      fpGitConf.write("\tAuthName \"%s Git Repository\"\n" % (name.replace('"',r'\"')))

      if apache_2_2:
         fpGitConf.write("\tAuthzLDAPAuthoritative on\n")
      if ldapbinduser != "":
         fpGitConf.write("\tAuthLDAPBindDN \"%s\"\n" % (ldapbinduser))
         fpGitConf.write("\tAuthLDAPBindPassword \"%s\"\n" % (ldapbindpass))
      if ldapgroupobjectclass == 'hubGroup':
         fpGitConf.write("\tAuthLDAPGroupAttributeIsDN on\n")
         fpGitConf.write("\tAuthLDAPGroupAttribute owner\n")
         fpGitConf.write("\tAuthLDAPGroupAttribute member\n")
      else:
         fpGitConf.write("\tAuthLDAPGroupAttributeIsDN off\n")
         fpGitConf.write("\tAuthLDAPGroupAttribute memberUid\n")

      fpGitConf.write("\tAuthLDAPURL %s/ou=users,%s\n" % (ldaphost,ldapbase))

      for group in gid:
         if ldapgroupobjectclass == 'hubGroup':
            fpGitConf.write("\tRequire ldap-group gid=%s,ou=groups,%s\n" % (group,ldapbase))
         else:
            fpGitConf.write("\tRequire ldap-group cn=%s,ou=groups,%s\n" % (group,ldapbase))
      fpGitConf.write("\tSatisfy Any\n")

      fpGitConf.write("</LocationMatch>\n")
      fpGitConf.write("\n")

   fpGitConf.close()

db.close()

